from typing import Any

from itertools import ifilter as filter
from itertools import imap as map
from itertools import izip as zip


def ascii(obj: Any) -> str: ...


def hex(x: int) -> str: ...


def oct(x: int) -> str: ...
